<?php
/**
 * Created by PhpStorm.
 * User: eater
 * Date: 1/17/18
 * Time: 11:54 AM
 */

namespace BitCommunism\Http;


use DI\Container;
use FastRoute\Dispatcher;
use FastRoute\RouteCollector;
use function FastRoute\simpleDispatcher;
use function GuzzleHttp\Psr7\copy_to_stream;
use GuzzleHttp\Psr7\Response;
use GuzzleHttp\Psr7\ServerRequest;
use function GuzzleHttp\Psr7\stream_for;
use Psr\Http\Message\RequestInterface;
use Psr\Http\Message\ResponseInterface;
use Psr\Http\Message\ServerRequestInterface;

class Server
{
    private $routes = [];
    private $container;

    /**
     * @var Dispatcher
     */
    private $router;

    public function __construct($routes, Container $container)
    {
        $this->routes = $routes;
        $this->container = $container;
    }

    private function iterateRoutes()
    {
        yield from $this->iterateRoutesFromArray($this->routes);
    }

    private function iterateRoutesFromArray($routeArr, $prefix = '', $known = [], $middleware = [])
    {
        foreach ($routeArr as $key => $route) {
            if ($key === 'middleware') {
                $middleware = array_merge($middleware, $route);
            }

            if (is_array($route) && !(isset($route[0]) && isset($route[1]))) {
                yield from $this->iterateRoutesFromArray($route, $prefix . $key, $known);
                continue;
            }

            if (is_string($route)) {
                $route = [$route, 'handle', []];
            }

            $route[2] = array_merge($middleware, $route[2] ?? []);

            if ($key[0] !== '/') {
                yield [$key, $prefix, $route];

                continue;
            }

            yield ['GET', $prefix . $key, $route];
        }
    }


    public function buildRoutes()
    {
        $this->router = simpleDispatcher(function (RouteCollector $collector) {
            foreach ($this->iterateRoutes() as $route) {
                $collector->addRoute(...$route);
            }
        });
    }

    public function serve()
    {
        $request = ServerRequest::fromGlobals();

        $response = $this->respond($request);

        header('HTTP/' . $response->getProtocolVersion() . ' ' . $response->getStatusCode() . ' ' . $response->getReasonPhrase());

        foreach ($response->getHeaders() as $headerName => $headerValues) {
            $first = true;
            foreach ($headerValues as $headerValue) {
                header($headerName . ': ' . $headerValue, $first === true);
                $first = false;
            }
        }

        $stdout = fopen('php://output', 'w+');

        copy_to_stream($response->getBody(), stream_for($stdout));
    }

    public function respond(ServerRequestInterface $request, ResponseInterface $response = null): ResponseInterface
    {
        $path = $request->getUri()->getPath();
        $method = $request->getMethod();
        $routeInfo = $this->router->dispatch($method, $path);

        switch ($routeInfo[0]) {
            case Dispatcher::NOT_FOUND:
                // todo: Error page
                break;
            case Dispatcher::METHOD_NOT_ALLOWED:
                // todo: Error page
                break;
            case Dispatcher::FOUND:
                [$_, $handler, $vars] = $routeInfo;
                $response = $this->handleHttp($request, $handler, $vars, $response);
                break;
        }

        if ($response === null) {
            // todo: Error page
        }

        return $response;
    }

    public function handleHttp(ServerRequestInterface $request, array $handler, array $vars, ResponseInterface $response = null): ResponseInterface
    {
        [$_, $_, $middlewareArray] = $handler;

        $middlewarePointer = 0;

        $next = null;

        if ($response === null) {
            $response = new Response(200, [], stream_for('No answer'));
        }

        $next = function ($request, $response, $vars) use ($handler, $middlewareArray, &$middlewarePointer, &$next) {
            $output = null;

            if (count($middlewareArray) >= $middlewarePointer) {
                $obj = $this->container->get($handler[0]);
                $call = [$obj, $handler[1]];
            } else {
                [$name, $func] = $middlewareArray[$middlewarePointer++];
                $middleware = $this->container->get($name);
                $call = [$middleware, $func];
            }

            $output = $this->container->call($call, [
                RequestInterface::class => $request,
                ResponseInterface::class => $response,
                'request' => $request,
                'response' => $response,
                'vars' => $vars,
            ]);


            return $this->normalizeResponse($output, $response);
        };

        return $next($request, $response, $vars);
    }

    private function normalizeResponse($response, ResponseInterface $previousResponse)
    {
        if ($response instanceof ResponseInterface) {
            return $response;
        }

        if (!$response) {
            return $previousResponse;
        }

        $body = stream_for($response);
        return $previousResponse->withBody($body);
    }
}